<?php declare(strict_types=1);
// Autor (c) Miroslav Novak, www.platiti.cz
// Pouzivani bez souhlasu autora neni povoleno
// #Ver:PRV108-60-g39033a71:2024-08-14#


namespace Platiti\UniAdapter\Service;

use Platiti\PtiUniNs\UniModul\UniErr;
use Shopware\Core\Framework\Context;
use Shopware\Core\Framework\DataAbstractionLayer\EntityRepository;
use Shopware\Core\Framework\DataAbstractionLayer\Search\Criteria;
use Shopware\Core\Framework\DataAbstractionLayer\Search\Filter\EqualsFilter;
use Shopware\Core\Framework\Plugin;
use Shopware\Core\Framework\Plugin\Context\ActivateContext;
use Shopware\Core\Framework\Plugin\Context\DeactivateContext;
use Shopware\Core\Framework\Plugin\Context\InstallContext;
use Shopware\Core\Framework\Plugin\Context\UninstallContext;
use Shopware\Core\Framework\Plugin\Util\PluginIdProvider;

if (!defined('E_UNIERR_DEFAULT')) {
	define('E_UNIERR_DEFAULT', E_ALL & ~E_STRICT & ~E_USER_DEPRECATED);
}

require_once(__DIR__ . '/../UniModul/UniModul.php');

class UniPlugin extends Plugin
{
	
	/** @var string */
	protected $uniModulName;

	public function install(InstallContext $context): void
	{
		$this->addPaymentMethod($context->getContext());
	}

	public function uninstall(UninstallContext $context): void
	{
		UniErr::run(null, function() use ($context) {
			// Only set the payment method to inactive when uninstalling. Removing the payment method would
			// cause data consistency issues, since the payment method might have been used in several orders
			$this->setPaymentMethodIsActive(false, $context->getContext());
		});
	}

	public function activate(ActivateContext $context): void
	{
		UniErr::run(null, function() use ($context) {
			$this->setPaymentMethodIsActive(true, $context->getContext());
			parent::activate($context);
		});
	}

	public function deactivate(DeactivateContext $context): void
	{
		UniErr::run(null, function() use ($context) {
			$this->setPaymentMethodIsActive(false, $context->getContext());
			parent::deactivate($context);
		});
	}

	private function addPaymentMethod(Context $context): void
	{
		$paymentMethodExists = $this->getPaymentMethodId();

		// Payment method exists already, no need to continue here
		if ($paymentMethodExists) {
			return;
		}

		/** @var PluginIdProvider $pluginIdProvider */
		$pluginIdProvider = $this->container->get(PluginIdProvider::class);
		$pluginId = $pluginIdProvider->getPluginIdByBaseClass(get_class($this), $context);

		$paymentData = [
			// payment handler will be selected by the identifier
			'handlerIdentifier' => $this->getPaymentPluginClassName(),
			'name' => $this->uniModulName,
			//'description' => $this->uniModulName . 'Platiti payment',
			'pluginId' => $pluginId,
		];

		/** @var EntityRepository $paymentRepository */
		$paymentRepository = $this->container->get('payment_method.repository');
		$paymentRepository->create([$paymentData], $context);
	}

	private function setPaymentMethodIsActive(bool $active, Context $context): void
	{
		/** @var EntityRepository $paymentRepository */
		$paymentRepository = $this->container->get('payment_method.repository');

		$paymentMethodId = $this->getPaymentMethodId();

		// Payment does not even exist, so nothing to (de-)activate here
		if (!$paymentMethodId) {
			return;
		}

		$paymentMethod = [
			'id' => $paymentMethodId,
			'active' => $active,
		];

		$paymentRepository->update([$paymentMethod], $context);
	}

	private function getPaymentMethodId(): ?string
	{
		/** @var EntityRepository $paymentRepository */
		$paymentRepository = $this->container->get('payment_method.repository');

		// Fetch ID for update
		$paymentCriteria = (new Criteria())->addFilter(new EqualsFilter('handlerIdentifier', $this->getPaymentPluginClassName()));
		return $paymentRepository->searchIds($paymentCriteria, Context::createDefaultContext())->firstId();
	}

	private function getPaymentPluginClassName(): string
	{
		return 'Platiti\\'.$this->uniModulName.'\\Service\\'.$this->uniModulName . 'Payment';
		//return Platiti\TestModul\Service\TestModulPayment::class;
	}
}
