<?php
// Autor (c) Miroslav Novak, www.platiti.cz
// Pouzivani bez souhlasu autora neni povoleno
// #Ver:PRV108-60-g39033a71:2024-08-14#

namespace Platiti\PtiUniNs\UniModul;

use DateTime;
use DateTimeInterface;
use Exception;
use const Platiti\PtiUniNs\UniModul\ConstUniModulVarPath;

if (!defined('Platiti\\PtiUniNs\\UniModul\\ConstUniModulVarPath')) define('Platiti\\PtiUniNs\\UniModul\\ConstUniModulVarPath', __DIR__); // kontanta ConstUniModulVarPath jen pro statické účely logování před vznikem UniModulu, všude jinde používat UniModul->baseConfig->uniModulVarPath

require_once('DatabaseConnection.php');
require_once('UniStructures.php');
require_once('UniDebug.php');
require_once('UniSync.php');
require_once('Session.php');
require_once('Http.php');
require_once('WebTools.php');

class UniModulConfig {
	/** @var ?string */
	var $mysql_server;
	/** @var ?string */
	var $mysql_dbname;
	/** @var ?string */
	var $mysql_login;
	/** @var ?string */
	var $mysql_password;
	/** @var ?IDatabaseConnection */
	var $databaseConnection;  // IDatabaseConnection, if it is set, the field mysql_* are not used
	/** @var string */
	var $uniModulDirUrl;
	/** @var callable(string, array<string, string>) : string */
	var $funcGetCallbackUrl;
	/** @var callable(OrderReplyStatus) : void */
	var $funcProcessReplyStatus; //urceno typicky pro davkove parovani transakci z banky atp., jinak pouzit standardni reply a notification
	/** @var string */
	var $adapterName;
	/** @var SessionInterface */
	public $session;
	/** @var string */
	public  $uniModulVarPath = ConstUniModulVarPath;
	/** @var array<int, string> */
	public  $legacyUniModulVarPaths = [];
	/** @var UniLogger */
	public $logger;
	/** @var ?int */
	public $uniShopId;
	/** @var array<int, string> */
	public  $adapterCapabilities = [];
}

abstract class UniModul {
	/** @var string */
	var $name;
	/** @var UniLogger */
	var $logger;
	/** @var WebTools */
	var $webTools;
	/** @var UniModulConfig */
	var $baseConfig;
	/** @var object */
	public $config; //public kvuli mocku
	/** @var UniDictionary */
	public $dictionary;
	/** @var ?string */
	public $subMethod = null; // 1 - pokud adapter dokaze v jednom modulu zobrazit vice metod, jmeno submetody je-li pro adapter znama jiz pri konstrukci, null nebo "" pro defaultni modul
	/** @var IDatabaseConnection */
	protected $dbConn;
	/** @var callable(UniModul) : bool */
	public $vrfacke = 'is_null';
	/** @var ?int */
	protected $uniShopId = null;
	/** @var array<string, mixed> */
	public $uniModulProperties = array(); // asoc pole vlastnosti/pozadavku UniModulu
	/** @var string Version with git hash PRV000-999-gffffffff from $versionStr */
	public $version;
	/** @var string Version git commit date yyyy-mm-dd parsed from $versionStr */
	public $versionDate;
	/** @var string Version number 000.999 separated by dot parsed from $versionStr*/
	public $versionDotNum;
	/** @var string Module version from GIT #Ver:PRV000-999-gffffffff:yyyy-mm-dd#*/
	public $versionStr = "#Ver:PRV108-60-g39033a71:2024-08-14#";
	/** @var string */
	var $configAcke;
	/** @var bool */
	var $activationKeyOk;


	/**
	 * @param string $name
	 * @param ConfigSetting $configSetting
	 * @param ?string $subMethod
	 * @return void
	 */
	public function __construct($name, $configSetting, $subMethod) {
		$this->name = $name;
		$this->subMethod = $subMethod;
		preg_match("/^#Ver:(.+):(.+)#/", $this->versionStr, $verpars);
		$this->version = isset($verpars[1]) ? $verpars[1] : 'DEVV';
		$this->versionDate = isset($verpars[2]) ? $verpars[2] : 'DEVT';
		if ($configSetting !== null && $configSetting->uniModulConfig->logger !== null) {
			$this->logger = $configSetting->uniModulConfig->logger;
		} else {
			$this->logger = UniLogger::getLogger();
		}
		$this->versionDotNum = preg_replace("/\D*0*(\d+)-(\d+)-.*/", "$1.$2", $this->version) ?: $this->version;
		$this->webTools = new WebTools($this->logger);
		if ($configSetting!==null) {
			$this->baseConfig = $configSetting->uniModulConfig;
			if ($this->baseConfig->databaseConnection != null) {
				$this->dbConn = $this->baseConfig->databaseConnection;
			} else if ($this->baseConfig->mysql_server !== null && $this->baseConfig->mysql_login  !== null && $this->baseConfig->mysql_dbname !== null) {
				$this->dbConn = new MysqliConnectionShared($this->baseConfig->mysql_server, $this->baseConfig->mysql_login, (string)$this->baseConfig->mysql_password, $this->baseConfig->mysql_dbname);
			} else if ($this->baseConfig->mysql_server !== null || $this->baseConfig->mysql_login  !== null || $this->baseConfig->mysql_dbname !== null) {
				throw new Exception("Missing database connection parameters");
			} // else connection not created, various get config info etc
			$this->checkConfig($configSetting);
			$this->uniShopId = $this->baseConfig->uniShopId;
		}
		$this->dictionary = $this->getDictionary();
	}

	/** @return UniDictionary  */
	protected function getDictionary() {
		return new UniDictionary($this->name, $this->logger);
	}


	/**
	 * @param string $language
	 * @return ConfigInfo
	 */
	abstract public function getConfigInfo($language='en'); //vola to jen UniFactory
	//vyreseno, neni nutne soucast interfacu//  abstract public function setConfigFromData($configSetting);  //vola to konstruktor odvozeneho UniModulu, nekdy ale kdyz konfigurace neni k mani pri konstrukci adapteru, tak se to volat jeste dodatecne

	// vrati mozne subMetody formou Array of strings
	/** @return array<int, string>  */
	public function getSubMethods() {
		return array();
	}


	//// detail produktu na splatky
	/**
	 * @param ?string $shopBaseUrl - deprecated, use baseConfig->uniModulDirUrl or getModulSubMethodLogoImage2
	 * @param string $currency
	 * @param float $price
	 * @param string $language
	 * @return ?string
	 */
	public function ProductGetInstallmentEmbedHtml($shopBaseUrl, $currency, $price, $language='cs') {
		return null;
	}

	//// volane pri placeni
	/**
	 * @param OrderToPayInfo $orderToPayInfo
	 * @return PrePayGWInfo
	 */
	abstract public function queryPrePayGWInfo($orderToPayInfo);
	/**
	 * @param OrderToPayInfo $orderToPayInfo
	 * @return RedirectAction
	 */
	abstract public function gatewayOrderRedirectAction($orderToPayInfo);
	/**
	 * @param string $language
	 * @return OrderReplyStatus
	 */
	abstract public function gatewayReceiveReply($language='en');

	/**
	 * @return OrderReplyStatus
	 */
	public function gatewayReceiveNotification() {
		throw new Exception('Not implemented');
	}

	/**
	 * @param OrderReplyStatus $orderReplyStatus
	 * @param string $shopOrderNumber
	 * @return void
	 */
	public function updateShopOrderNumber($orderReplyStatus, $shopOrderNumber) {
		$sql = "update unimodul_transactions set dateModified = FROM_UNIXTIME(".time()."), shopOrderNumber=".toSql($shopOrderNumber)." where gwOrderNumber = ".toSql($orderReplyStatus->gwOrderNumber)." and uniModulName=".toSql($this->name) . $this->getUniShopIdWhere();
		$this->dbConn->sqlExecute($sql);
	}




	/**
	 * @param ?string $gwOrderNumber
	 * @param ?int $transactionPK
	 * @param ?string $gwPairingInfo
	 * @return null|TransactionRecord
	 *
	 */
	public function getOrderTransactionRecordFromDbUnique($gwOrderNumber, $transactionPK=null, $gwPairingInfo=null) {
		return $this->getOrderTransactionRecordFromDbPriv(null, $gwOrderNumber, null, $transactionPK, $gwPairingInfo)[0];
	}

	/**
	 * @param ?string $gwOrderNumber
	 * @param ?int $transactionPK
	 * @param ?string $gwPairingInfo
	 * @return TransactionRecord
	 *
	 */
	public function getOrderTransactionRecordFromDbUniqueOrFail($gwOrderNumber, $transactionPK=null, $gwPairingInfo=null) {
		list($transactionRecord, $sql) = $this->getOrderTransactionRecordFromDbPriv(null, $gwOrderNumber, null, $transactionPK, $gwPairingInfo);
		if ($transactionRecord === null) {
			throw new Exception("Transaction not found, query: \"$sql\"");
		}
		return $transactionRecord;
	}

	// pri prohlizeni obejdnavky napr v prehledu objednavek v eshopu pro zjisteni napr. c.obj. na brane
	/**
	 * @param ?string $shopOrderNumber
	 * @param ?string $shopPairingInfo
	 * @return null|TransactionRecord
	 */
	public function getOrderTransactionRecordFromDbLast($shopOrderNumber, $shopPairingInfo=null) {
		return $this->getOrderTransactionRecordFromDbPriv($shopOrderNumber, null, $shopPairingInfo, null)[0];
	}

	/**
	 * @param ?string $shopOrderNumber
	 * @param ?string $shopPairingInfo
	 * @return TransactionRecord
	 */
	public function getOrderTransactionRecordFromDbLastOrFail($shopOrderNumber, $shopPairingInfo=null) {
		list($transactionRecord, $sql) = $this->getOrderTransactionRecordFromDbPriv($shopOrderNumber, null, $shopPairingInfo, null);
		if ($transactionRecord === null) {
			throw new Exception("Transaction not found, query: \"$sql\"");
		}
		return $transactionRecord;
	}

	/**
	 * @param ?string $shopOrderNumber
	 * @param ?string $gwOrderNumber
	 * @param ?string $shopPairingInfo
	 * @param ?int $transactionPK
	 * @param ?string $gwPairingInfo
	 * @return array{null|TransactionRecord, string}
	 */
	private function getOrderTransactionRecordFromDbPriv($shopOrderNumber, $gwOrderNumber=null, $shopPairingInfo=null, $transactionPK=null, $gwPairingInfo=null) {
		if (!is_null($transactionPK)) {
			$where = "transactionPK=".toSql($transactionPK);
		} else if (!is_null($gwOrderNumber)) {
			$where = "gwOrderNumber=".toSql($gwOrderNumber);
		} else if (!is_null($gwPairingInfo)) {
			$where = "gwPairingInfo=".toSql($gwPairingInfo);
		} else if (!is_null($shopPairingInfo)) {
			$where = "shopPairingInfo=".toSql($shopPairingInfo);
		} else if (!is_null($shopOrderNumber)) {
			$where = "shopOrderNumber=".toSql($shopOrderNumber);
		} else {
			throw new Exception("At least one identifier has to be provided");
		}
		$sql = "select * from unimodul_transactions where ".$where." and uniModulName=".toSql($this->name) . $this->getUniShopIdWhere() . " order by if(orderStatus=0, 99, orderStatus), transactionPK desc";
		$ar = $this->dbConn->sqlQuery($sql);

		if (count($ar)==0) { // to je potreba napr pro Prestu, kde v adminu hookovane infa o gwOrdnumberu se volaji vsechny moduly, i kdyz nemaji nic spolecneho s danou platbou
			return [null, $sql];
		}
		if (count($ar)>1 && ($gwOrderNumber !== null || $transactionPK !== null)) {
			$this->logger->writeLog("ERROR UniModul: duplikatni zaznamy v unimodul_transaction: ".$sql);
		}
		return [$this->loadTransactionRecordFromDbRow($ar[0]), $sql];
	}


	/**
	 * @param array<string, mixed> $row
	 * @return TransactionRecord
	 */
	protected function loadTransactionRecordFromDbRow($row) {
		$transactionRecord = new TransactionRecord();
		$transactionRecord->transactionPK =intval($row['transactionPK']);
		$transactionRecord->uniModulName = strval($row['uniModulName']);
		$transactionRecord->gwOrderNumber = strvalNull($row['gwOrderNumber']);
		$transactionRecord->gwPairingInfo = strvalNull($row['gwPairingInfo']);
		$transactionRecord->shopOrderNumber = strvalNull($row['shopOrderNumber']);
		$transactionRecord->shopPairingInfo = strval($row['shopPairingInfo']);
		$transactionRecord->uniAdapterData = unserialize(strval($row['uniAdapterData']));
		$transactionRecord->uniModulData = unserialize(strval($row['uniModulData']));
		$transactionRecord->forexNote = strvalNull($row['forexNote']);
		$transactionRecord->orderStatus = intval($row['orderStatus']);
		$transactionRecord->dateCreated = new DateTime(strval($row['dateCreated']));
		$transactionRecord->dateModified = $row['dateModified'] === null ? null : new DateTime(strval($row['dateModified']));
		return $transactionRecord;
	}

	// rozhrani pro zdedene UniModuly  - protected

	/**
	 * @param ?string $shopOrderNumber
	 * @param string $shopPairingInfo
	 * @param ?string $gwOrderNumber
	 * @param ?string $forexNote
	 * @param int $orderStatus
	 * @param array<string, mixed> $uniAdapterData
	 * @param array<string, mixed> $uniModulData
	 * @param ?string $gwPairingInfo
	 * @return int
	 */
	protected function writeOrderToDb($shopOrderNumber, $shopPairingInfo, $gwOrderNumber, $forexNote, $orderStatus=null, $uniAdapterData=null, $uniModulData=null, $gwPairingInfo=null) {
		if ($orderStatus === null) {
			$orderStatus = OrderStatus::$initiated;
		}
		list($spcol, $spval) = $this->getUniShopIdInsert();
		$sql = "insert into unimodul_transactions (".$spcol."uniModulName, gwOrderNumber, shopOrderNumber, shopPairingInfo, forexNote, orderStatus, uniAdapterData, uniModulData, gwPairingInfo, dateCreated) values ("
		.$spval
		.toSql($this->name).", "
		.toSql($gwOrderNumber).","
		.toSql($shopOrderNumber).","
		.toSql($shopPairingInfo).","
		.toSql($forexNote).","
		.$orderStatus.","
		.toSql(serialize($uniAdapterData)).","
		.toSql(serialize($uniModulData)).","
		.toSql($gwPairingInfo).","
		."FROM_UNIXTIME(".time().")"
		.")";
		$this->dbConn->sqlExecute($sql);
		$insId = $this->dbConn->getInsertId();
		return $insId;
	}


	/**
	 * @param int $transactionPK
	 * @param string $gwOrderNumber
	 * @param string $gwPairingInfo
	 * @return void
	 */
	protected function updateGwOrderNumber($transactionPK, $gwOrderNumber, $gwPairingInfo = null) {
		if ($gwPairingInfo !== null) {
			$pinf = ", gwPairingInfo=".toSql($gwPairingInfo) . " ";
		} else {
			$pinf = "";
		}
		$sql = "update unimodul_transactions set dateModified = FROM_UNIXTIME(".time()."), gwOrderNumber=".toSql($gwOrderNumber) . $pinf . " where transactionPK = ".toSql($transactionPK) . $this->getUniShopIdWhere();
		$this->dbConn->sqlExecute($sql);
	}

	/**
	 * @param OrderReplyStatus $orderReplyStatus
	 * @param ?int $transactionPK
	 * @return void
	 */
	protected function updateOrderReplyStatusGwOrdNumInDb($orderReplyStatus, $transactionPK = null) {
		if ($transactionPK !== null) {
			$sql = "update unimodul_transactions set DateModified = FROM_UNIXTIME(".time()."), orderStatus=".toSql($orderReplyStatus->orderStatus).", gwOrderNumber=".toSql($orderReplyStatus->gwOrderNumber).", gwPairingInfo=".toSql($orderReplyStatus->gwPairingInfo)." where transactionPK = ".toSql($transactionPK) . $this->getUniShopIdWhere();
			$this->dbConn->sqlExecute($sql);
		} else if ($orderReplyStatus->gwOrderNumber != null) {
			$sql = "update unimodul_transactions set DateModified = FROM_UNIXTIME(".time()."), orderStatus=".toSql($orderReplyStatus->orderStatus).", gwPairingInfo=".toSql($orderReplyStatus->gwPairingInfo)." where gwOrderNumber = ".toSql($orderReplyStatus->gwOrderNumber)." and uniModulName=".toSql($this->name) . $this->getUniShopIdWhere();
			$this->dbConn->sqlExecute($sql);
		} else {
			throw new Exception("updateOrderReplyStatusGwOrdNumInDb chybi transactionId i orderReplyStatus->gwOrderNumber");
		}
	}

	/**
	 * @param OrderReplyStatus $orderReplyStatus
	 * @return int
	 */
	protected function ensureGlobalPairingInfoStatusUpgradeOnly($orderReplyStatus) {
		// overime ze nedojde ke snizeni urovne dokonceni objednavky podle pairingInfo, aby se nezmenil stav v eshopu po te co nektery z predchozich pokusu o zaplaceni zpozdene zahlasi neuspech
		if ($orderReplyStatus->orderStatus == OrderStatus::$successful || $orderReplyStatus->orderStatus == OrderStatus::$invalidReply) {
			$hledStat = null;
		} else if ($orderReplyStatus->orderStatus == OrderStatus::$pending) {
			$hledStat = OrderStatus::$successful;
		} else {
			$hledStat = OrderStatus::$successful . ',' . OrderStatus::$pending;
		}
		if ($hledStat !== null) {
			$sql = "select * from unimodul_transactions where orderStatus in (".$hledStat.") and shopPairingInfo=".toSql($orderReplyStatus->shopPairingInfo) . " and gwOrderNumber != ".toSql($orderReplyStatus->gwOrderNumber) . $this->getUniShopIdWhere();
			$ar = $this->dbConn->sqlQuery($sql);
		} else {
			$ar = array();
		}
		if (count($ar) != 0) {
			$this->logger->writeLog("WARNING: Jiz existuje stav platba objednavky s vyssi urovni zaplaceni, prepiname na invalidReply. Status stavajici objednavky je ".$orderReplyStatus->orderStatus.", ale existujici obejdnavka GwOrdNum=" . $ar[0]['gwOrderNumber'] . " ma stav " . $ar[0]['orderStatus']);
			return OrderStatus::$invalidReply;
		} else {
			return $orderReplyStatus->orderStatus;
		}
	}

	/**
	 * @param int $transactionPK
	 * @param array<string, mixed> $uniAdapterData
	 * @return void
	 */
	public function updateUniAdapterDataInDb($transactionPK, $uniAdapterData) {
		$sql = "update unimodul_transactions set dateModified = FROM_UNIXTIME(".time()."), uniAdapterData=".toSql(serialize($uniAdapterData))." where transactionPK = ".toSql($transactionPK) . $this->getUniShopIdWhere();
		$this->dbConn->sqlExecute($sql);
	}

	/**
	 * @param int $transactionPK
	 * @param array<string, mixed> $uniModulData
	 * @return void
	 */
	public function updateUniModulDataInDb($transactionPK, $uniModulData) {
		$sql = "update unimodul_transactions set dateModified = FROM_UNIXTIME(".time()."), uniModulData=".toSql(serialize($uniModulData))." where transactionPK = ".toSql($transactionPK) . $this->getUniShopIdWhere();
		$this->dbConn->sqlExecute($sql);
	}


	// zjisteni pending transakci pri offline zjistovani stavu transakce z cronu
	/**
	 * @param ?array<int, int> $statuses
	 * @param ?DateTimeInterface $since
	 * @return array<TransactionRecord>
	 */
	protected function getAllPendingOrderTransactionRecords($statuses = null, $since = null) {
		if ($statuses === null) $statuses = array(OrderStatus::$pending);
		array_walk($statuses, function (&$s) {$s = toSql($s);});
		$dateCond = '';
		if ($since !== null) {
			$dateCond = ' and dateCreated >= ' . toSql(date_format($since,"Y-m-d H:i:s"));
		}
		$sql = "select * from unimodul_transactions where uniModulName=".toSql($this->name)." and orderStatus in (".implode(",", $statuses).") " . $dateCond . $this->getUniShopIdWhere() . "order by greatest(dateCreated, dateModified)";
		$ar = $this->dbConn->sqlQuery($sql);
		$pendingTransactions = array();
		foreach($ar as $ar0) {
			$transactionRecord = $this->loadTransactionRecordFromDbRow($ar0);
			$pendingTransactions[]=$transactionRecord;
		}
		return $pendingTransactions;
	}

	/** @return string  */
	protected function getUniShopIdWhere() {
		if ($this->uniShopId !==null) {
			return " and uniShopId=".toSql($this->uniShopId)." ";
		} else {
			return "";
		}
	}

	/** @return array{string, string}  */
	protected function getUniShopIdInsert() {
		if ($this->uniShopId !==null) {
			return array("uniShopId,", toSql($this->uniShopId) . ",");
		} else {
			return array("", "");
		}
	}


	/**
	 * @template _T
	 * @param callable(): _T $func
	 * @return _T
	 */
	public function runSynchronized($func) {
		try {
			$lockName = "UniModul_Lock_" . $this->uniShopId;
			$r = $this->dbConn->sqlQuery("select get_lock(" . toSql($lockName) . ", 0) as r")[0]['r'];
			if (!$r) {
				$this->logger->writeLog("INFO: Sync BLOCK: Waiting for lock");
				$r = $this->dbConn->sqlQuery("select get_lock(" . toSql($lockName) . ", 180) as r")[0]['r'];
				if ($r) {
					$this->logger->writeLog("INFO: Sync BLOCK: Lock acquired");
				} else {
					$this->logger->writeLog("ERROR: Sync BLOCK: Lock acquisition failed - continuing without lock");
				}
			}
			return $func();
		} finally {
			$r = $this->dbConn->sqlQuery("select release_lock(" . toSql($lockName) . ") as r")[0]['r'];
			if (!$r) {
				$this->logger->writeLog("ERROR: Lock release failed");
			}
		}
	}

	/**
	 * @param OrderToPayInfo $orderToPayInfo    mutates all prices to the new currency if neccessary
	 * @param ?bool $currencySupported
	 * @param bool $allowZeroAmount
	 * @return array{bool, string, float, ?string, ?string, ?OrderReplyStatus}
	 */
	protected function fixCurrency($orderToPayInfo, $currencySupported = null, $allowZeroAmount = false) {
		if ($currencySupported===null && isset($this->config->supportedCurrencies)) {
			$currencySupported = empty($orderToPayInfo->currency) ? false : strpos($this->config->supportedCurrencies, $orderToPayInfo->currency)!==false;
		}
		$forexMessage = null;
		$forexNote = null;
		$newtotal = $orderToPayInfo->amount;
		$actcur = $orderToPayInfo->currency;
		$newcur = $actcur;
		$newtotal = $orderToPayInfo->amount;
		$orderReplyStatusFail = null;
		if ($newtotal == 0 && !$allowZeroAmount) {
			$isPossible = false;
			$orderReplyStatusFail = $this->createImmediateReplyStatus($orderToPayInfo, OrderStatus::$failedRetriable, "Zero amount");
		} else if (!$currencySupported) {
			if (!empty($this->config->convertToCurrencyIfUnsupported) && $orderToPayInfo->currencyRates !== null) {
				$newcur = strval($this->config->convertToCurrencyIfUnsupported);
				$rate = $orderToPayInfo->currencyRates[$newcur] / $orderToPayInfo->currencyRates[$actcur];
				$newtotal = $orderToPayInfo->amount * $rate;

				$orderToPayInfo->amount = $newtotal;
				$orderToPayInfo->currency = $newcur;
				if ($orderToPayInfo->cartItems != null) {
					array_walk($orderToPayInfo->cartItems, function($v, $k) use ($rate) {$v->unitPrice *= $rate;});
				}

				$ratestr = number_format($rate, 4, '.', ' ');
				$newtotalstr = number_format($newtotal, 2, '.', ' ');

				$forexMessageTemplate = $this->dictionary->get('forexMessageTemplate', $orderToPayInfo->language);
				$forexMessage = strtr($forexMessageTemplate, array(
					"{actcur}"=>$actcur,
					"{newcur}"=>$newcur,
					"{newtotalstr}"=>$newtotalstr,
					"{ratestr}"=>$ratestr,
				));

				$forexNote = "{$orderToPayInfo->amount} $actcur -> $newtotalstr $newcur @ $ratestr $newcur/$actcur";
				if (!empty($GLOBALS['UniModul_TrackDisplayPaymentOption'])) {
					$this->logger->writeLog("fixCurrency " . $forexNote);
				}
				$isPossible = true;
			} else {
				$isPossible = false;

				$resultText = $this->dictionary->get('unsupportedCurrency', $orderToPayInfo->language)." ".$orderToPayInfo->currency;
				$orderReplyStatusFail = $this->createImmediateReplyStatus($orderToPayInfo, OrderStatus::$failedRetriable, $resultText);
			}
		} else {
			$isPossible = true;
		}
		$ok = call_user_func($this->vrfacke, $this);
		$isPossible = $isPossible && $ok;

		if (!empty($GLOBALS['UniModul_TrackDisplayPaymentOption'])) {
			$this->logger->writeLogNoNewLines("fixCurrency (".$this->name." ".$this->subMethod."): isPossible = " . ($isPossible?"true":"false") . "   currencySupported:".($currencySupported?"true":"false")  . "  orderToPayInfo:".print_r($orderToPayInfo, true) . "  " . ($ok?"ACK":" NAK"));
		}
		if (!$isPossible && $orderReplyStatusFail === null) {
			$orderReplyStatusFail = $this->createImmediateReplyStatus($orderToPayInfo, OrderStatus::$failedRetriable);
		}
		return array($isPossible, $newcur, $newtotal, $forexMessage, $forexNote, $orderReplyStatusFail);
	}

	/**
	 * @param OrderToPayInfo $orderToPayInfo
	 * @param int $orderStatus
	 * @param ?string $resultText
	 * @param ?string $successHtml
	 * @return OrderReplyStatus
	 */
	function createImmediateReplyStatus($orderToPayInfo, $orderStatus, $resultText = null, $successHtml = null) {
		$orderReplyStatusFail = new OrderReplyStatus();
		$orderReplyStatusFail->orderStatus = $orderStatus;
		$orderReplyStatusFail->resultText = $resultText;
		$orderReplyStatusFail->successHtml = $successHtml;
		$orderReplyStatusFail->shopOrderNumber = $orderToPayInfo->shopOrderNumber;
		//$orderReplyStatusFail->gwOrderNumber = $gwOrderNumber;
		$orderReplyStatusFail->shopPairingInfo = $orderToPayInfo->shopPairingInfo;
		$orderReplyStatusFail->uniAdapterData = $orderToPayInfo->uniAdapterData;
		return $orderReplyStatusFail;
	}

	/**
	 * @param OrderToPayInfo $orderToPayInfo
	 * @param int $orderStatus
	 * @param ?string $resultText
	 * @param ?string $successHtml
	 * @return RedirectAction
	 */
	public function getImmediateReplyStatusRedirectAction($orderToPayInfo, $orderStatus, $resultText = null, $successHtml = null) {
		$redirectAction = new RedirectAction();
		$redirectAction->orderReplyStatus = $this->createImmediateReplyStatus($orderToPayInfo, $orderStatus, $resultText, $successHtml);
		return $redirectAction;
	}

	// infoBox

	/**
	 * @param string $uniAdapterName
	 * @param string $language
	 * @return InfoBoxData
	 */
	public function getInfoBoxData($uniAdapterName, $language) {
		$infoBoxLinkTextTemplate = $this->dictionary->get('infoBoxPlatitiLinkText', $language);
		$linkText = strtr($infoBoxLinkTextTemplate, array(
			"{modulname}"=>$this->name,
			"{shopname}"=>$uniAdapterName,
		));

		$infoBoxData = new InfoBoxData();
		$infoBoxData->title = $this->name;
		$infoBoxData->image = "Uni".$this->name."Logo.png";
		//$infoBoxData->link = "http://www.payu.cz";  // dosadi se az v konkretnim modulu
		$infoBoxData->platitiLink = "http://www.platiti.cz/{$uniAdapterName}-{$this->name}.php";
		$infoBoxData->platitiLinkText = $linkText;
		return $infoBoxData;
	}

	// helpers


	/**
	 * @param string $form
	 * @return never
	 * @deprecated use $this->webTools
	 */
	public function formRedirect($form) {
		$this->webTools->formRedirect($form);
	}

	/**
	 * @param string $redirUrl
	 * @return never
	 * @deprecated use $this->webTools
	 */
	public function jsRedirect($redirUrl) {
		$this->webTools->jsRedirect($redirUrl);
	}

	/**
	 * @return bool
	 * @deprecated use $this->webTools
	 */
	public function isSafariBrowser() {
		return $this->webTools->isSafariBrowser();
	}

	/**
	 * @param ConfigSetting $configSetting
	 * @return void
	 */
	function checkConfig($configSetting) {
		if ($configSetting != null && $configSetting->configData != null) {
			$this->configAcke = strval($configSetting->configData['activationKey']);
			$chksumok = true;
			foreach(explode(' ',$this->configAcke) as $akey0) {
				$akey = str_replace('-', '', $akey0);
				if (strlen($akey)<2) continue;
				if (substr(md5(substr($akey,0,-2)), 0,2) != substr($akey,-2)) {
					$chksumok = false;
					$this->logger->writeLog("Wro"."ng key " . $akey0);
				}
			}
			$this->activationKeyOk = $chksumok;
			$this->vrfacke = __NAMESPACE__.'\\vrfacke'; /** @phpstan-ignore-line */
			if (!function_exists(__NAMESPACE__.'\\vrfacke')) {
				function vrfacke ($ths) { /** @phpstan-ignore-line */
					$dom = $GLOBALS["_SER"."VER"]["HTT"."P_H"."OST"];
					$actKeys = $ths->configAcke;
					$mod = $ths->name;
					$ada = $ths->baseConfig->adapterName;
					$upddatee = explode('-', $ths->versionDate);
					$upddate15 = count($upddatee)>2 ? (intval($upddatee[0])-2015)*12+(intval($upddatee[1])-1) : 0;

					$ok = false;
					/** @var string */
					$dom = preg_replace('/(:[0-9]+)/', '', $dom);
					$dom = strtolower($dom);
					/** @var string */
					$dom3 = preg_replace('/(^www\.)/', '', $dom);
					if ($dom3=='localhost' || $dom3=='127.0.0.1') {
						$ok = true;
					}
					foreach(explode(' ',$actKeys) as $actkey) {
						$actkey = str_replace('-', '', $actkey);
						/** @var string */
						$actkey = preg_replace('/[^0-9a-f]/','0',$actkey);
						$actkey = str_pad($actkey, 14, '0');

						$rndh = substr($actkey, 0, 2);
						$datesh = substr($actkey, 2, 6);
						$dates = intval(base_convert($datesh, 16, 10)) ^ intval(base_convert($rndh.$rndh.$rndh, 16, 10));
						$maxdate = intval($dates>>12);
						$maxupddate = $dates % (1<<12);
						$date15 = (date("y")-15)*12+(date("m")-1);
						$dateok = $date15<=$maxdate && $upddate15 <= $maxupddate;

						$numhash = intval(base_convert(substr(md5($dom3.$ada.$mod.$datesh.$rndh),0, 4), 16, 10)) % 61423;
						$numhashdev = intval(base_convert(substr(md5($dom3.$datesh.$rndh),0, 4), 16, 10)) % 61423;
						/** @psalm-suppress ArgumentTypeCoercion */
						$dec = \bcpowmod(base_convert(substr($actkey,8,4), 16, 10), strval(43), strval(61423), 0);
						$domok = $numhash == $dec || $numhashdev == $dec;

						$ok = $ok || ($dateok && $domok);
					}
					if (!$ok) {
						$ths->logger->writeLog("Ba"."d key $ada $mod $dom '$actKeys'");
					}
					return $ok;
				}
			}

		}
	}

	/**
	 * @param string $language
	 * @param ?string $subMethod
	 * @return string
	 */
	public function getModuleSubMethodName($language='en', $subMethod='') {
		if ($subMethod == '') $subMethod = $this->subMethod;
		if ($subMethod == '') {
			$subname = $this->dictionary->get('payment_method_name', $language);
		} else {
			$subname = $this->dictionary->get('submethod_name_'.$subMethod, $language);
		}
		return $subname;
	}

	/**
	 * @param string $unused_shopBaseUrl
	 * @param string $subMethod
	 * @return string
	 * @deprecated use getModulSubMethodLogoImage2 instead
	 */
	public function getModulSubMethodLogoImage($unused_shopBaseUrl, $subMethod='') {
		return $this->getModulSubMethodLogoImage2($subMethod);
	}

	/**
	 * @param ?string $subMethod
	 * @return string
	 */
	public function getModulSubMethodLogoImage2($subMethod='') {
		if ($subMethod == '') $subMethod = $this->subMethod;
	    if (file_exists(dirname(__FILE__)."/Uni".$this->name.$subMethod."Logo.png")) {
	    	$submetLogoPiece = $subMethod;
	    } else {
	    	$submetLogoPiece = "";
	    }
	    $logoimg = $this->baseConfig->uniModulDirUrl . "Uni".urlencode($this->name . $submetLogoPiece)."Logo.png";
	    return $logoimg;
	}


	/** @return string  */
	public function getMiniLogoUrl() {
		return "https://www.platiti.cz/muzo/minilogo.png";
	}

	/** @return string  */
	public function getMiniLogoSpan() {
		return "<span style=\"background:url('".$this->getMiniLogoUrl()."')\"></span>";
	}

	/**
	 * @param string $callbackName
	 * @param array<string, string> $arguments
	 * @return void
	 */
	public function processCallbackRequest($callbackName, $arguments) {
		$this->logger->writeLog("Cannot process callback request " . Http::server('REQUEST_URI'));
		throw new Exception("Cannot process callback request");
	}

	// dopocita nepovinne polozky v cartItems
	/**
	 * @param OrderToPayInfo $orderToPayInfo
	 * @return void
	 */
	public function normalizeCartItems($orderToPayInfo /*updatuje*/) {
		$vychoziOrderToPayInfo = unserialize(serialize($orderToPayInfo));
		foreach($orderToPayInfo->cartItems as $item) {
			if (empty($item->type)) {
				$itemName = (string)$item->name;
				if ($itemName == 'shipping / doprava') $item->type = UniCartItemType::delivery;
				else if ($itemName == 'discount / sleva') $item->type = UniCartItemType::discount;
				else $item->type = UniCartItemType::commodity;
			}
			if ($item->unitTaxAmount === null) {
				$item->unitTaxAmount = $item->unitPrice * (1 - 100/(100 + $item->taxRate));
			}
		}
	}


	// sesumiruje polozkyp pres jednotlive sazby
	/**
	 * @param OrderToPayInfo $orderToPayInfo
	 * @return array{array<int, float>, array<int, float>}	vraci [array<sazba, totalAmount>, array<sazba, totalTax>]
	 */
	public function getTaxRozdeleni($orderToPayInfo) {
		// secteni pres sazby
		$rozdeleniAmount = array();
		$rozdeleniTax = array();
		foreach ($orderToPayInfo->cartItems as $ci) {
			$ciTaxRateIdx = (int)round($ci->taxRate);

			$act = $ci->unitPrice * $ci->quantity;
			$cum = isset($rozdeleniAmount[$ciTaxRateIdx]) ? $rozdeleniAmount[$ciTaxRateIdx] : 0;
			$rozdeleniAmount[$ciTaxRateIdx] = $cum + $act;

			if ($ci->unitTaxAmount !== null) {
				$act = $ci->unitTaxAmount * $ci->quantity;
			} else {
				$koef = $ci->taxRate/(100 + $ci->taxRate);
				$act = $ci->unitPrice * $koef  * $ci->quantity;
			}
			$cum = isset($rozdeleniTax[$ciTaxRateIdx]) ? $rozdeleniTax[$ciTaxRateIdx] : 0;
			$rozdeleniTax[$ciTaxRateIdx] = $cum + $act;
		}
		return [$rozdeleniAmount, $rozdeleniTax];
	}

	// sesumiruje polozkyp pres jednotlive sazby a provede korekci zaokrouhleni, vysledek EetCastkySazby
	/**
	 * @param OrderToPayInfo $orderToPayInfo
	 * @param bool $convertToCzk
	 * @return null|EetCastkySazby
	 */
	public function getEetRozdeleni($orderToPayInfo, $convertToCzk = true) {

		if (empty($orderToPayInfo->cartItems)) {
			$this->logger->writeLog("Prazdne cart Items");
			return null;
		}

		list($rozdeleni, $rozdeleniTax) = $this->getTaxRozdeleni($orderToPayInfo);

		// prevod do CZK
		if ($convertToCzk && $orderToPayInfo->currency != 'CZK' && isset($orderToPayInfo->currencyRates['CZK'])) {
			$fxrate = $orderToPayInfo->currencyRates['CZK'] / $orderToPayInfo->currencyRates[$orderToPayInfo->currency];
			array_walk($rozdeleni, function(&$v, $k) use ($fxrate) {$v = $v * $fxrate;});
			array_walk($rozdeleniTax, function(&$v, $k) use ($fxrate) {$v = $v * $fxrate;});
			$total =$orderToPayInfo->amount * $fxrate;
		} else {
			$total =$orderToPayInfo->amount;
		}


		// zaokrouhleni na halere
		array_walk($rozdeleni, function(&$v, $k) {$v = round($v, 2);});
		array_walk($rozdeleniTax, function(&$v, $k) {$v = round($v, 2);});

		// korekce pro pripadne zaokrouhleni do nejnizsi ze sazeb
		$total = round($total, 2);
		$vychoziSuma = array_sum($rozdeleni);
		$rozdilZaokrouhleni = $total - $vychoziSuma;
		$minsazba = array_reduce(array_keys($rozdeleni), 'min', 000000);
		$rozdeleni[$minsazba] += $total - $vychoziSuma;
		if (abs($rozdilZaokrouhleni) > 10) {
			$this->logger->writeLog("Nesouhlasi castky v polozkach kosiku. Placena castka " . $total . ", soucet polozek je " . $vychoziSuma . "Polozky kosiku: " . print_r($orderToPayInfo->cartItems, true));
			return null;
		}

		// kontrola na pripustne sazby pro DPH
		$eetSazby = array(0, 21, 15, 10);
		$sazbyNok = array_diff(array_keys($rozdeleni), $eetSazby);
		if (count($sazbyNok)>0) {
			$this->logger->writeLog("Neplatna sazba DPH " . implode(",", $sazbyNok) . ", polozky kosiku: " . print_r($orderToPayInfo->cartItems, true));
			return null;
		}


		// vypocet zakladu a dane
		$eetCastky = new EetCastkySazby();
		$eetCastky->celk_trzba = array_sum($rozdeleni);
		if (isset($rozdeleni[0])) {
			$eetCastky->zakl_nepodl_dph = $rozdeleni[0];
		}
		for ($i=1; $i<=3; $i++) {
			$sazba = $eetSazby[$i];
			if (isset($rozdeleni[$sazba])) {
				$celk = $rozdeleni[$sazba];
				if (empty($rozdeleniTax[$sazba])) {
					$koef = round($sazba/(100 + $sazba), 4);
					$dan = round($celk * $koef, 2);
				} else {
					$dan = $rozdeleniTax[$sazba];
				}
				$eetCastky->{"zakl_dan$i"} = $celk - $dan;
				$eetCastky->{"dan$i"} = $dan;
			}
		}
		return $eetCastky;
	}

	// vraci poznamku pro pridani k objednavce nebo null
	/**
	 * @param ?string $shopOrderNumber
	 * @param string $shopPairingInfo
	 * @param string $newShopOrderStatus
	 * @param ?callable() : OrderToPayInfo $funcGetOrderToPayInfo
	 * @return null | array{success: bool, message: string, newStatus?: string, result?: string}
	 */
	public function shopOrderStatusChanged($shopOrderNumber, $shopPairingInfo, $newShopOrderStatus, $funcGetOrderToPayInfo = null) {
		return null;
	}

	/**
	 * @param ?string $shopOrderNumber
	 * @param string $shopPairingInfo
	 * @return ?string
	 */
	public function getOrderAdminDetailHtml($shopOrderNumber, $shopPairingInfo) {
		return null;
	}

	/**
	 * @param UniLogger $logger
	 * @return void
	 */
	public static function moduleInstallVarDirStatic($logger) {

		$md = function ($d) {
			$p = ConstUniModulVarPath . '/' . $d;
			if (is_dir($p)) return true;
			return mkdir($p, 0777, true);
		};

		$ok = $md('logs')
		&& file_put_contents(ConstUniModulVarPath . '/logs/.htaccess', "deny from all\n")
		&& $md('certs')  // TODO: toto by melo byt jen u modulu, ktere certifikaty potrebuji
		&& file_put_contents(ConstUniModulVarPath . '/certs/.htaccess', "deny from all\n")
		&& touch(ConstUniModulVarPath . '/sync.lock')
		&& copy(__DIR__ . '/UniModulConfig.php.example', ConstUniModulVarPath . '/UniModulConfig.php.example')
		&& copy(__DIR__ . '/version.txt', ConstUniModulVarPath . '/version.txt');
		if (!$ok) throw new Exception("ERROR UniModul: Install procedure cannot create UniModulVar directory tree in " . ConstUniModulVarPath);
		$logger->writeLog("INFO UniModul: Install - Provided var directory structure in ". ConstUniModulVarPath);
	}

	/**
	 * @return void
	 */
	public function moduleInstallVarDir() {
		static::moduleInstallVarDirStatic($this->logger);
	}


	/**
	 * @param IDatabaseConnection $dbconn
	 * @param UniLogger $logger
	 * @return void
	 */
	public static function moduleInstallDatabaseStatic($dbconn, $logger) {
		$logger->writeLog("INFO UniModul: Install - Providing db table unimodul_transactions");
		$dbconn->sqlExecute((string)file_get_contents( __DIR__ . "/UniModul.sql"));

		$sql = "SELECT CHARACTER_MAXIMUM_LENGTH FROM INFORMATION_SCHEMA.COLUMNS WHERE TABLE_SCHEMA=DATABASE() AND TABLE_NAME='unimodul_transactions' AND column_name='gwOrderNumber'";
		if ($dbconn->sqlQuery($sql)[0]['CHARACTER_MAXIMUM_LENGTH'] < 100) {
			$sql = '
				alter table unimodul_transactions
				modify gwOrderNumber varchar(100) DEFAULT NULL;
			';
			$logger->writeLog("INFO UniModul: Install - fixing db table structure unimodul_transactions: " . $sql);
			$dbconn->sqlExecute($sql);
		}

		$sql = "SELECT count(*) as present FROM INFORMATION_SCHEMA.COLUMNS WHERE TABLE_SCHEMA=DATABASE() AND TABLE_NAME='unimodul_transactions' AND column_name='gwPairingInfo'";
		if ($dbconn->sqlQuery($sql)[0]['present'] == 0) {
			$sql = '
				alter table unimodul_transactions
				add column gwPairingInfo varchar(80) DEFAULT NULL,
				add column gwAccount varchar(80) DEFAULT NULL;
			';
			$logger->writeLog("INFO UniModul: Install - fixing db table structure unimodul_transactions: " . $sql);
			$dbconn->sqlExecute($sql);
		}
}

	/**
	 * @return void
	 */
	public function moduleInstallDatabase() {
		static::moduleInstallDatabaseStatic($this->dbConn,$this->logger);
	}

}

class EetCastkySazby {
	/** @var float */
	var $celk_trzba;
	/** @var float */
	var $zakl_nepodl_dph;
	/** @var float */
	var $zakl_dan1;
	/** @var float */
	var $dan1;
	/** @var float */
	var $zakl_dan2;
	/** @var float */
	var $dan2;
	/** @var float */
	var $zakl_dan3;
	/** @var float */
	var $dan3;
}

class UniModulFactory {
	/**
	 * @param string $name
	 * @param string $language
	 * @param ?string $subMethod
	 * @param string $adapterName
	 * @return ConfigInfo
	 */
	function getConfigInfo($name, $language='en', $subMethod=null, $adapterName=null) {
		$uniModulConfig = create_initialize_object(UniModulConfig::class, array('adapterName' => $adapterName));
		return $this->getConfigInfo2($name, $uniModulConfig, $language);
	}

	/**
	 * @param string $name
	 * @param string $language
	 * @param UniModulConfig $uniModulConfig
	 * @return ConfigInfo
	 */
	function getConfigInfo2($name, $uniModulConfig, $language='en') {
		$preCofingInfo = create_initialize_object(ConfigSetting::class, ['uniModulConfig' => $uniModulConfig]);
		$uniModul = $this->createUniModul($name, $preCofingInfo);
		$cofingInfo = $uniModul->getConfigInfo($language);

		$d = $uniModul->dictionary;
		$d->setDefaultLanguage($language);
		$configField = new ConfigField();
		$configField->name = 'activationKey';
		$configField->label = $d->get('activationKey');
		$configField->type = ConfigFieldType::$text;
		array_unshift($cofingInfo->configFields, $configField);

		return $cofingInfo;
	}

	/**
	 * @param string $name
	 * @param ConfigSetting $configSetting
	 * @param ?string $subMethod
	 * @return UniModul
	 */
	function createUniModul($name, $configSetting, $subMethod=null) {
		if (!ctype_alnum($name)) {
			throw new Exception('Neplatny UniModul: '.$name);
		}

		if ($configSetting->uniModulConfig->session === null) {
			$configSetting->uniModulConfig->session = new StandardSession();
		}


		require_once("Uni".$name.".php");
		$vniModul = "\\Platiti\PtiUniNs\\" . $name . "\\Uni" . $name;
		if (!class_exists($vniModul)) {
			$vniModul = "Uni" . $name;
		}
		/** @var UniModul */
		$unimod = new $vniModul($configSetting, $subMethod);
		return $unimod;
	}
}



class UniDictionary {
	/** @var array<string, array<string, string>> */
	protected $dictionary;
	/** @var UniLogger */
	protected $logger;
	/** @var string */
	protected $dictionaryFile;

	/**
	 * @param string $dictionaryFile
	 * @param UniLogger $logger
	 * @return void
	 */
	function __construct($dictionaryFile, $logger) {
		/** @var array<string, array<string, string>> */
		$dict = array();
		include(dirname(__FILE__).'/Uni'."Modul"."Lang.php");  // nejak parametrizovat aby to bylo pripravene pro pluginy
		include(dirname(__FILE__).'/Uni'.$dictionaryFile."Lang.php");
		$this->dictionary = $dict;
		$this->dictionaryFile = $dictionaryFile;
		$this->logger = $logger;
		/* pro nacitani z CSV
		$fh = fopen(dirname(__FILE__).'/'.$dictionaryFile, "r");
			$langs = fgetcsv($fh, 1000, ";");
			$this->dictionary = array();
			for ($i=1; $i<count($langs); $i++) {
				$this->dictionary[$langs[$i]]=array();
			}
			while ($line = fgetcsv($fh, 1000, ";")) {
				$key = $line[0];
				for ($i=1; $i<count($line); $i++) {
					$this->dictionary[$langs[$i]][$key] = $line[$i];
				}
			}
		*/
	}

	/** @var string */
	private $defaultLanguage;

	/**
	 * @param string $language
	 * @return void
	 */
	function setDefaultLanguage($language) {
		$this->defaultLanguage = $language;
	}

	/**
	 * @param string $key
	 * @param string $language
	 * @return string
	 */
	function get($key, $language=null) {
		if ($language === null) {
			$language = $this->defaultLanguage;
		}
		if (isset($this->dictionary[$language]) && isset($this->dictionary[$language][$key])) {
			return $this->dictionary[$language][$key];
		} else {
			if (isset($this->dictionary['en'][$key])) {
				$r = $this->dictionary['en'][$key];
			} else {
				$r = $key;
			}
			//$this->logger->writeLog("WARN UniModul: Missing translation [$language:$key] in {$this->dictionaryFile} using '$r'");
			//user_error("WARN UniModul: Missing translation [$language:$key] in {$this->dictionaryFile} using $r");
			return $r;
		}
	}


}



abstract class UniLogger {

	/** @var UniLogger */
	protected static $logger;

	/** @return UniLogger  */
	public static function getLogger() {
		if (self::$logger == null) {
			self::$logger = new UniLoggerClassic();
		}
		return self::$logger;
	}

	/**
	 * @param UniLogger $logger
	 * @return void
	 */
	public static function setLogger($logger) {
		self::$logger = $logger;
	}


	/**
	 * @param string $s
	 * @return void
	 */
	 abstract public function writeLog($s);
	/**
	 * @param string $s
	 * @return void
	 */
	public function writeLogNoNewLines($s) {
		$this->writeLog($this->replaceLogCrLf($s));
	}

	/**
	 * @param string $s
	 * @return string
	 */
	public function replaceLogCrLf($s) {
		$s = str_replace("\r","",str_replace("\n","|",$s));
		return $s;
	}
}



class UniLoggerClassic extends UniLogger {
	/** @var string */
	var $logFile;

	function __construct() {
		$this->logFile = isset($GLOBALS['UniModul_Logfile']) ? $GLOBALS['UniModul_Logfile'] : ConstUniModulVarPath . "/logs/UniModul.log";
		if (!file_exists(dirname($this->logFile))) {
			mkdir(dirname($this->logFile), 0777, true);
		}
	}

	public function writeLog($s) {
		$line = "*** ".date('r')." ".$s."\n";
		file_put_contents($this->logFile, $line, FILE_APPEND | LOCK_EX);
	}

}




/**
 * @template _T
 * @param class-string<_T> $className
 * @param array<string, mixed> $data
 * @return _T
 */
function create_initialize_object($className, $data) { // data array(n=>v)
	$object = new $className();
	foreach($data as $name => $value) {
		$object->{$name} = $value;
	}
	return $object;
}

/**
 * @param mixed $val
 * @return null|string
 */
function strvalNull($val) {
	return $val === null ? null : strval($val);
}


if ("PtiUniNs" == "Pti"."UniNs") require_once('UniModulNoNs.php');  // @phpstan-ignore-line
